// Copyright 2023 Georges Racinet <georges.racinet@octobus.net>
//
// This software may be used and distributed according to the terms of the
// GNU General Public License version 2 or any later version.
// SPDX-License-Identifier: GPL-2.0-or-later
use std::fmt::{Debug, Formatter};
use std::string::String;
use std::sync::Arc;
use tonic::{metadata::Ascii, metadata::MetadataValue, Request, Response, Status};
use tracing::{field, info, instrument};

use crate::config::Config;
use crate::gitaly::ref_service_server::{RefService, RefServiceServer};
use crate::gitaly::{
    FindDefaultBranchNameRequest, FindDefaultBranchNameResponse, RefExistsRequest,
    RefExistsResponse,
};
use crate::gitlab::gitlab_branch_ref;
use crate::gitlab::revision::{existing_default_gitlab_branch, map_full_ref, RefError};
use crate::metadata::correlation_id;
use crate::repository::{default_repo_spec_error_status, repo_store_vfs};

#[derive(Debug)]
pub struct RefServiceImpl {
    config: Arc<Config>,
}

#[tonic::async_trait]
impl RefService for RefServiceImpl {
    async fn find_default_branch_name(
        &self,
        request: Request<FindDefaultBranchNameRequest>,
    ) -> Result<Response<FindDefaultBranchNameResponse>, Status> {
        let (metadata, _ext, inner) = request.into_parts();

        self.inner_find_default_branch_name(inner, correlation_id(&metadata))
            .await
    }
    async fn ref_exists(
        &self,
        request: Request<RefExistsRequest>,
    ) -> Result<Response<RefExistsResponse>, Status> {
        let (metadata, _ext, inner) = request.into_parts();

        self.inner_ref_exists(inner, correlation_id(&metadata))
            .await
    }
}

struct RefExistsTracingRequest<'a>(&'a RefExistsRequest);

impl Debug for RefExistsTracingRequest<'_> {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("RefExistRequest")
            .field("repository", &self.0.repository)
            .field("ref", &String::from_utf8_lossy(&self.0.r#ref))
            .finish()
    }
}

impl RefServiceImpl {
    #[instrument(name = "find_default_branch_name", skip(self))]
    async fn inner_find_default_branch_name(
        &self,
        request: FindDefaultBranchNameRequest,
        correlation_id: Option<&MetadataValue<Ascii>>,
    ) -> Result<Response<FindDefaultBranchNameResponse>, Status> {
        info!("Processing");

        let store_vfs = repo_store_vfs(&self.config, &request.repository)
            .await
            .map_err(default_repo_spec_error_status)?;

        Ok(Response::new(FindDefaultBranchNameResponse {
            name: existing_default_gitlab_branch(&store_vfs)
                .await
                .map_err(|e| {
                    Status::internal(format!(
                        "Error reading or checking GitLab default branch: {:?}",
                        e
                    ))
                })?
                .map(|ref name_node| gitlab_branch_ref(&name_node.0))
                .unwrap_or_else(Vec::new),
        }))
    }

    #[instrument(name="ref_exists",
                 skip(self, request),
                 fields(request=field::debug(RefExistsTracingRequest(&request))))]
    async fn inner_ref_exists(
        &self,
        request: RefExistsRequest,
        correlation_id: Option<&MetadataValue<Ascii>>,
    ) -> Result<Response<RefExistsResponse>, Status> {
        info!("Processing");

        let store_vfs = repo_store_vfs(&self.config, &request.repository)
            .await
            .map_err(default_repo_spec_error_status)?;

        let value = match map_full_ref(&store_vfs, &request.r#ref, |_tr| (), |_ka| ()).await {
            Ok(()) => Ok(true),
            Err(RefError::NotFound) => Ok(false),
            Err(RefError::MissingRefName) => Ok(false),
            Err(RefError::NotAFullRef) => Err(Status::invalid_argument("invalid refname")),
            Err(RefError::GitLabStateFileError(e)) => Err(Status::internal(format!("{:?}", e))),
        }?;

        Ok(Response::new(RefExistsResponse { value }))
    }
}

/// Takes care of boilerplate that would instead be in the startup sequence.
pub fn ref_server(config: &Arc<Config>) -> RefServiceServer<RefServiceImpl> {
    RefServiceServer::new(RefServiceImpl {
        config: config.clone(),
    })
}
