// Copyright 2023 Georges Racinet <georges.racinet@octobus.net>
//
// This software may be used and distributed according to the terms of the
// GNU General Public License version 2 or any later version.
// SPDX-License-Identifier: GPL-2.0-or-later
use std::sync::Arc;
use tokio_stream::StreamExt;
use tonic::{metadata::Ascii, metadata::MetadataValue, Request, Response, Status};
use tracing::{info, instrument};

use crate::config::Config;
use crate::gitaly::repository_service_server::{RepositoryService, RepositoryServiceServer};
use crate::gitaly::{
    HasLocalBranchesRequest, HasLocalBranchesResponse, RepositoryExistsRequest,
    RepositoryExistsResponse,
};
use crate::gitlab::state::stream_gitlab_branches;
use crate::metadata::correlation_id;
use crate::repository::{
    checked_repo_path, default_repo_spec_error_status, repo_store_vfs, RepoSpecError,
};

#[derive(Debug)]
pub struct RepositoryServiceImpl {
    config: Arc<Config>,
}

#[tonic::async_trait]
impl RepositoryService for RepositoryServiceImpl {
    async fn repository_exists(
        &self,
        request: Request<RepositoryExistsRequest>,
    ) -> Result<Response<RepositoryExistsResponse>, Status> {
        let (metadata, _ext, inner) = request.into_parts();

        self.inner_repository_exists(inner, correlation_id(&metadata))
            .await
    }

    async fn has_local_branches(
        &self,
        request: Request<HasLocalBranchesRequest>,
    ) -> Result<Response<HasLocalBranchesResponse>, Status> {
        let (metadata, _ext, inner) = request.into_parts();

        self.inner_has_local_branches(inner, correlation_id(&metadata))
            .await
            .map(|v| Response::new(HasLocalBranchesResponse { value: v }))
    }
}

impl RepositoryServiceImpl {
    #[instrument(name = "repository_exists", skip(self))]
    async fn inner_repository_exists(
        &self,
        request: RepositoryExistsRequest,
        correlation_id: Option<&MetadataValue<Ascii>>,
    ) -> Result<Response<RepositoryExistsResponse>, Status> {
        info!("Processing");

        match checked_repo_path(&self.config, &request.repository).await {
            Ok(_) => Ok(true),
            Err(RepoSpecError::RepoNotFound(_)) => Ok(false),
            Err(e) => Err(default_repo_spec_error_status(e)),
        }
        .map(|res| Response::new(RepositoryExistsResponse { exists: res }))
    }

    #[instrument(name = "has_local_branches", skip(self))]
    async fn inner_has_local_branches(
        &self,
        request: HasLocalBranchesRequest,
        correlation_id: Option<&MetadataValue<Ascii>>,
    ) -> Result<bool, Status> {
        info!("Processing");

        let store_vfs = repo_store_vfs(&self.config, &request.repository)
            .await
            .map_err(default_repo_spec_error_status)?;
        if let Some(mut stream) = stream_gitlab_branches(&store_vfs).await.map_err(|e| {
            Status::internal(format!("Problem reading Gitlab branches file: {:?}", e))
        })? {
            Ok(stream.next().await.is_some())
        } else {
            Ok(false)
        }
    }
}

/// Takes care of boilerplate that would instead be in the startup sequence.
pub fn repository_server(config: &Arc<Config>) -> RepositoryServiceServer<RepositoryServiceImpl> {
    RepositoryServiceServer::new(RepositoryServiceImpl {
        config: config.clone(),
    })
}
